@@ -39,12 +39,16 @@
 #include <stdbool.h>
 #include <string.h>
 #include <inttypes.h>
-#include <alloca.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
+#include <assert.h>
+
+#ifdef HAVE_ALLOCA_H
+#include <alloca.h>
+#endif
 
 #include <nbdkit-filter.h>
 
@@ -55,6 +59,8 @@
 
 /* Size of a block in the cache.  A 4K block size means that we need
  * 64 MB of memory to store the bitmaps for a 1 TB underlying image.
+ * It is also smaller than the usual hole size for sparse files, which
+ * means we have no reason to call next_ops->zero.
  */
 #define BLKSIZE 4096
 
@@ -72,7 +78,7 @@ static int fd = -1;
 static uint8_t *bitmap;
 
 /* Size of the bitmap in bytes. */
-static uint64_t bm_size;
+static size_t bm_size;
 
 enum bm_entry {
   BLOCK_NOT_CACHED = 0,
@@ -87,6 +93,10 @@ static enum cache_mode {
   CACHE_MODE_UNSAFE,
 } cache_mode = CACHE_MODE_WRITEBACK;
 
+static int
+cache_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle,
+             uint32_t flags, int *err);
+
 static void
 cache_load (void)
 {
@@ -167,7 +177,14 @@ blk_set_size (uint64_t new_size)
 {
   uint8_t *new_bm;
   const size_t old_bm_size = bm_size;
-  size_t new_bm_size = DIV_ROUND_UP (new_size, BLKSIZE*8/2);
+  uint64_t new_bm_size_u64 = DIV_ROUND_UP (new_size, BLKSIZE*8/2);
+  size_t new_bm_size;
+
+  if (new_bm_size_u64 > SIZE_MAX) {
+    nbdkit_error ("bitmap too large for this architecture");
+    return -1;
+  }
+  new_bm_size = (size_t) new_bm_size_u64;
 
   new_bm = realloc (bitmap, new_bm_size);
   if (new_bm == NULL) {
@@ -179,7 +196,7 @@ blk_set_size (uint64_t new_size)
   if (old_bm_size < new_bm_size)
     memset (&bitmap[old_bm_size], 0, new_bm_size-old_bm_size);
 
-  nbdkit_debug ("cache: bitmap resized to %" PRIu64 " bytes", new_bm_size);
+  nbdkit_debug ("cache: bitmap resized to %zu bytes", new_bm_size);
 
   if (ftruncate (fd, new_size) == -1) {
     nbdkit_error ("ftruncate: %m");
@@ -218,7 +235,10 @@ cache_prepare (struct nbdkit_next_ops *next_ops, void *nxdata,
   int64_t r;
 
   r = cache_get_size (next_ops, nxdata, handle);
-  return r >= 0 ? 0 : -1;
+  if (r < 0)
+    return -1;
+  /* TODO: cache per-connection FUA mode? */
+  return 0;
 }
 
 /* Return true if the block is allocated.  Consults the bitmap. */
@@ -256,7 +276,7 @@ blk_set_bitmap_entry (uint64_t blknum, enum bm_entry state)
  */
 static int
 blk_read (struct nbdkit_next_ops *next_ops, void *nxdata,
-          uint64_t blknum, uint8_t *block)
+          uint64_t blknum, uint8_t *block, int *err)
 {
   off_t offset = blknum * BLKSIZE;
   enum bm_entry state = blk_get_bitmap_entry (blknum);
@@ -269,9 +289,10 @@ blk_read (struct nbdkit_next_ops *next_ops, void *nxdata,
                 "unknown");
 
   if (state == BLOCK_NOT_CACHED) /* Read underlying plugin. */
-    return next_ops->pread (nxdata, block, BLKSIZE, offset);
+    return next_ops->pread (nxdata, block, BLKSIZE, offset, 0, err);
   else {                         /* Read cache. */
     if (pread (fd, block, BLKSIZE, offset) == -1) {
+      *err = errno;
       nbdkit_error ("pread: %m");
       return -1;
     }
@@ -282,7 +303,8 @@ blk_read (struct nbdkit_next_ops *next_ops, void *nxdata,
 /* Write to the cache and the plugin. */
 static int
 blk_writethrough (struct nbdkit_next_ops *next_ops, void *nxdata,
-                  uint64_t blknum, const uint8_t *block)
+                  uint64_t blknum, const uint8_t *block, uint32_t flags,
+                  int *err)
 {
   off_t offset = blknum * BLKSIZE;
 
@@ -291,11 +313,12 @@ blk_writethrough (struct nbdkit_next_ops *next_ops, void *nxdata,
                 blknum, (uint64_t) offset);
 
   if (pwrite (fd, block, BLKSIZE, offset) == -1) {
+    *err = errno;
     nbdkit_error ("pwrite: %m");
     return -1;
   }
 
-  if (next_ops->pwrite (nxdata, block, BLKSIZE, offset) == -1)
+  if (next_ops->pwrite (nxdata, block, BLKSIZE, offset, flags, err) == -1)
     return -1;
 
   blk_set_bitmap_entry (blknum, BLOCK_CLEAN);
@@ -306,12 +329,14 @@ blk_writethrough (struct nbdkit_next_ops *next_ops, void *nxdata,
 /* Write to the cache only. */
 static int
 blk_writeback (struct nbdkit_next_ops *next_ops, void *nxdata,
-               uint64_t blknum, const uint8_t *block)
+               uint64_t blknum, const uint8_t *block, uint32_t flags,
+               int *err)
 {
   off_t offset;
 
-  if (cache_mode == CACHE_MODE_WRITETHROUGH)
-    return blk_writethrough (next_ops, nxdata, blknum, block);
+  if (cache_mode == CACHE_MODE_WRITETHROUGH ||
+      (cache_mode == CACHE_MODE_WRITEBACK && (flags & NBDKIT_FLAG_FUA)))
+    return blk_writethrough (next_ops, nxdata, blknum, block, flags, err);
 
   offset = blknum * BLKSIZE;
 
@@ -320,6 +345,7 @@ blk_writeback (struct nbdkit_next_ops *next_ops, void *nxdata,
                 blknum, (uint64_t) offset);
 
   if (pwrite (fd, block, BLKSIZE, offset) == -1) {
+    *err = errno;
     nbdkit_error ("pwrite: %m");
     return -1;
   }
@@ -331,12 +357,15 @@ blk_writeback (struct nbdkit_next_ops *next_ops, void *nxdata,
 /* Read data. */
 static int
 cache_pread (struct nbdkit_next_ops *next_ops, void *nxdata,
-             void *handle, void *buf, uint32_t count, uint64_t offset)
+             void *handle, void *buf, uint32_t count, uint64_t offset,
+             uint32_t flags, int *err)
 {
   uint8_t *block;
 
+  assert (!flags);
   block = malloc (BLKSIZE);
   if (block == NULL) {
+    *err = errno;
     nbdkit_error ("malloc: %m");
     return -1;
   }
@@ -350,7 +379,7 @@ cache_pread (struct nbdkit_next_ops *next_ops, void *nxdata,
     if (n > count)
       n = count;
 
-    if (blk_read (next_ops, nxdata, blknum, block) == -1) {
+    if (blk_read (next_ops, nxdata, blknum, block, err) == -1) {
       free (block);
       return -1;
     }
@@ -369,16 +398,24 @@ cache_pread (struct nbdkit_next_ops *next_ops, void *nxdata,
 /* Write data. */
 static int
 cache_pwrite (struct nbdkit_next_ops *next_ops, void *nxdata,
-              void *handle, const void *buf, uint32_t count, uint64_t offset)
+              void *handle, const void *buf, uint32_t count, uint64_t offset,
+              uint32_t flags, int *err)
 {
   uint8_t *block;
+  bool need_flush = false;
 
   block = malloc (BLKSIZE);
   if (block == NULL) {
+    *err = errno;
     nbdkit_error ("malloc: %m");
     return -1;
   }
 
+  if ((flags & NBDKIT_FLAG_FUA) &&
+      next_ops->can_fua (nxdata) == NBDKIT_FUA_EMULATE) {
+    flags &= ~NBDKIT_FLAG_FUA;
+    need_flush = true;
+  }
   while (count > 0) {
     uint64_t blknum, blkoffs, n;
 
@@ -389,12 +426,12 @@ cache_pwrite (struct nbdkit_next_ops *next_ops, void *nxdata,
       n = count;
 
     /* Do a read-modify-write operation on the current block. */
-    if (blk_read (next_ops, nxdata, blknum, block) == -1) {
+    if (blk_read (next_ops, nxdata, blknum, block, err) == -1){
       free (block);
       return -1;
     }
     memcpy (&block[blkoffs], buf, n);
-    if (blk_writeback (next_ops, nxdata, blknum, block) == -1) {
+    if (blk_writeback (next_ops, nxdata, blknum, block, flags, err) == -1) {
       free (block);
       return -1;
     }
@@ -405,22 +442,33 @@ cache_pwrite (struct nbdkit_next_ops *next_ops, void *nxdata,
   }
 
   free (block);
+  if (need_flush)
+    return cache_flush (next_ops, nxdata, handle, 0, err);
   return 0;
 }
 
 /* Zero data. */
 static int
 cache_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
-            void *handle, uint32_t count, uint64_t offset, int may_trim)
+            void *handle, uint32_t count, uint64_t offset, uint32_t flags,
+            int *err)
 {
   uint8_t *block;
+  bool need_flush = false;
 
   block = malloc (BLKSIZE);
   if (block == NULL) {
+    *err = errno;
     nbdkit_error ("malloc: %m");
     return -1;
   }
 
+  flags &= ~NBDKIT_FLAG_MAY_TRIM; /* See BLKSIZE comment above. */
+  if ((flags & NBDKIT_FLAG_FUA) &&
+      next_ops->can_fua (nxdata) == NBDKIT_FUA_EMULATE) {
+    flags &= ~NBDKIT_FLAG_FUA;
+    need_flush = true;
+  }
   while (count > 0) {
     uint64_t blknum, blkoffs, n;
 
@@ -430,12 +478,12 @@ cache_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
     if (n > count)
       n = count;
 
-    if (blk_read (next_ops, nxdata, blknum, block) == -1) {
+    if (blk_read (next_ops, nxdata, blknum, block, err) == -1) {
       free (block);
       return -1;
     }
     memset (&block[blkoffs], 0, n);
-    if (blk_writeback (next_ops, nxdata, blknum, block) == -1) {
+    if (blk_writeback (next_ops, nxdata, blknum, block, flags, err) == -1) {
       free (block);
       return -1;
     }
@@ -445,18 +493,22 @@ cache_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
   }
 
   free (block);
+  if (need_flush)
+    return cache_flush (next_ops, nxdata, handle, 0, err);
   return 0;
 }
 
 /* Flush: Go through all the dirty blocks, flushing them to disk. */
 static int
-cache_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
+cache_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle,
+             uint32_t flags, int *err)
 {
   uint8_t *block = NULL;
   uint64_t i, j;
   uint64_t blknum;
   enum bm_entry state;
   unsigned errors = 0;
+  int tmp;
 
   if (cache_mode == CACHE_MODE_UNSAFE)
     return 0;
@@ -466,7 +518,7 @@ cache_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
    * to be sure.  Also we still need to issue the flush to the
    * underlying storage.
    */
-
+  assert (!flags);
   for (i = 0; i < bm_size; ++i) {
     if (bitmap[i] != 0) {
       /* The bitmap stores information about 4 blocks per byte,
@@ -480,6 +532,7 @@ cache_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
           if (!block) {
             block = malloc (BLKSIZE);
             if (block == NULL) {
+              *err = errno;
               nbdkit_error ("malloc: %m");
               return -1;
             }
@@ -487,8 +540,10 @@ cache_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
           /* Perform a read + writethrough which will read from the
            * cache and write it through to the underlying storage.
            */
-          if (blk_read (next_ops, nxdata, blknum, block) == -1 ||
-              blk_writethrough (next_ops, nxdata, blknum, block)) {
+          if (blk_read (next_ops, nxdata, blknum, block,
+                        errors ? &tmp : err) == -1 ||
+              blk_writethrough (next_ops, nxdata, blknum, block, 0,
+                                errors ? &tmp : err) == -1) {
             nbdkit_error ("cache: flush of block %" PRIu64 " failed", blknum);
             errors++;
           }
@@ -500,7 +555,7 @@ cache_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
   free (block);
 
   /* Now issue a flush request to the underlying storage. */
-  if (next_ops->flush (nxdata) == -1)
+  if (next_ops->flush (nxdata, 0, errors ? &tmp : err) == -1)
     errors++;
 
   return errors == 0 ? 0 : -1;
