@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013 Red Hat Inc.
+ * Copyright (C) 2013-2018 Red Hat Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -36,21 +36,34 @@
 #ifndef NBDKIT_PLUGIN_H
 #define NBDKIT_PLUGIN_H
 
-#include <stdarg.h>
-#include <stdint.h>
+#include <nbdkit-common.h>
 
-#define NBDKIT_THREAD_MODEL_SERIALIZE_CONNECTIONS     0
-#define NBDKIT_THREAD_MODEL_SERIALIZE_ALL_REQUESTS    1
-#define NBDKIT_THREAD_MODEL_SERIALIZE_REQUESTS        2
-#define NBDKIT_THREAD_MODEL_PARALLEL                  3
+#ifdef __cplusplus
+extern "C" {
+#endif
 
+/* By default, a plugin gets API version 1; but you may request
+ * version 2 prior to including this header */
+#ifndef NBDKIT_API_VERSION
 #define NBDKIT_API_VERSION                            1
+#elif (NBDKIT_API_VERSION - 0) < 1 || NBDKIT_API_VERSION > 2
+#error Unsupported API version
+#endif
 
 struct nbdkit_plugin {
+  /* Do not set these fields directly; use NBDKIT_REGISTER_PLUGIN.
+   * They exist so that we can support plugins compiled against
+   * one version of the header with a runtime compiled against a
+   * different version with more (or fewer) fields.
+   */
   uint64_t _struct_size;
   int _api_version;
   int _thread_model;
 
+  /* Plugins are responsible for these fields; see the documentation
+   * for semantics, and which fields are optional. New fields will
+   * only be added at the end of the struct.
+   */
   const char *name;
   const char *longname;
   const char *version;
@@ -73,25 +86,47 @@ struct nbdkit_plugin {
   int (*is_rotational) (void *handle);
   int (*can_trim) (void *handle);
 
+#if NBDKIT_API_VERSION == 1
   int (*pread) (void *handle, void *buf, uint32_t count, uint64_t offset);
   int (*pwrite) (void *handle, const void *buf, uint32_t count, uint64_t offset);
   int (*flush) (void *handle);
   int (*trim) (void *handle, uint32_t count, uint64_t offset);
+  int (*zero) (void *handle, uint32_t count, uint64_t offset, int may_trim);
+#else
+  int (*_pread_old) (void *, void *, uint32_t, uint64_t);
+  int (*_pwrite_old) (void *, const void *, uint32_t, uint64_t);
+  int (*_flush_old) (void *);
+  int (*_trim_old) (void *, uint32_t, uint64_t);
+  int (*_zero_old) (void *, uint32_t, uint64_t, int);
+#endif
 
-  /* int (*set_exportname) (void *handle, const char *exportname); */
+  int errno_is_preserved;
+
+  void (*dump_plugin) (void);
+
+  int (*can_zero) (void *handle);
+  int (*can_fua) (void *handle);
+#if NBDKIT_API_VERSION == 1
+  int (*_unused1) (void *, void *, uint32_t, uint64_t);
+  int (*_unused2) (void *, const void *, uint32_t, uint64_t, uint32_t);
+  int (*_unused3) (void *, uint32_t);
+  int (*_unused4) (void *, uint32_t, uint64_t, uint32_t);
+  int (*_unused5) (void *, uint32_t, uint64_t, uint32_t);
+#else
+  int (*pread) (void *handle, void *buf, uint32_t count, uint64_t offset,
+                uint32_t flags);
+  int (*pwrite) (void *handle, const void *buf, uint32_t count,
+                 uint64_t offset, uint32_t flags);
+  int (*flush) (void *handle, uint32_t flags);
+  int (*trim) (void *handle, uint32_t count, uint64_t offset, uint32_t flags);
+  int (*zero) (void *handle, uint32_t count, uint64_t offset, uint32_t flags);
+#endif
 };
 
-extern void nbdkit_error (const char *msg, ...)
-  __attribute__((format (printf, 1, 2)));
-extern void nbdkit_verror (const char *msg, va_list args);
-extern void nbdkit_debug (const char *msg, ...)
-  __attribute__((format (printf, 1, 2)));
-extern void nbdkit_vdebug (const char *msg, va_list args);
-
-extern char *nbdkit_absolute_path (const char *path);
-extern int64_t nbdkit_parse_size (const char *str);
+extern void nbdkit_set_error (int err);
 
 #define NBDKIT_REGISTER_PLUGIN(plugin)                                  \
+  NBDKIT_CXX_LANG_C                                                     \
   struct nbdkit_plugin *                                                \
   plugin_init (void)                                                    \
   {                                                                     \
@@ -101,4 +136,8 @@ extern int64_t nbdkit_parse_size (const char *str);
     return &(plugin);                                                   \
   }
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* NBDKIT_PLUGIN_H */
