@@ -39,11 +39,9 @@
 #include <string.h>
 #include <inttypes.h>
 #include <assert.h>
-#include <errno.h>
 
 #include <dlfcn.h>
 
-#include "nbdkit-filter.h"
 #include "internal.h"
 
 /* We extend the generic backend struct with extra fields relating
@@ -78,11 +76,12 @@ filter_free (struct backend *b)
    */
   lock_unload ();
 
-  debug ("%s: unload", f->filename);
+  debug ("%s: unload", f->name);
   if (f->filter.unload)
     f->filter.unload ();
 
-  dlclose (f->dl);
+  if (DO_DLCLOSE)
+    dlclose (f->dl);
   free (f->filename);
 
   unlock_unload ();
@@ -104,23 +103,15 @@ filter_thread_model (struct backend *b)
   return thread_model;
 }
 
-/* These are actually passing through to the final plugin, hence
- * the function names.
+/* This is actually passing the request through to the final plugin,
+ * hence the function name.
  */
-static int
-plugin_errno_is_preserved (struct backend *b)
-{
-  struct backend_filter *f = container_of (b, struct backend_filter, backend);
-
-  return f->backend.next->errno_is_preserved (f->backend.next);
-}
-
 static const char *
 plugin_name (struct backend *b)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
 
-  return f->backend.next->name (f->backend.next);
+  return f->backend.next->plugin_name (f->backend.next);
 }
 
 static const char *
@@ -181,7 +172,7 @@ filter_config (struct backend *b, const char *key, const char *value)
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
 
   debug ("%s: config key=%s, value=%s",
-         f->filename, key, value);
+         f->name, key, value);
 
   if (f->filter.config) {
     if (f->filter.config (next_config, f->backend.next, key, value) == -1)
@@ -204,7 +195,7 @@ filter_config_complete (struct backend *b)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
 
-  debug ("%s: config_complete", f->filename);
+  debug ("%s: config_complete", f->name);
 
   if (f->filter.config_complete) {
     if (f->filter.config_complete (next_config_complete, f->backend.next) == -1)
@@ -229,7 +220,7 @@ filter_open (struct backend *b, struct connection *conn, int readonly)
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
   void *handle;
 
-  debug ("%s: open readonly=%d", f->filename, readonly);
+  debug ("%s: open readonly=%d", f->name, readonly);
 
   if (f->filter.open) {
     handle = f->filter.open (next_open, &nxdata, readonly);
@@ -248,7 +239,7 @@ filter_close (struct backend *b, struct connection *conn)
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
   void *handle = connection_get_handle (conn, f->backend.i);
 
-  debug ("close");
+  debug ("%s: close", f->name);
 
   if (f->filter.close)
     f->filter.close (handle);
@@ -298,43 +289,58 @@ next_can_trim (void *nxdata)
 }
 
 static int
-next_pread (void *nxdata, void *buf, uint32_t count, uint64_t offset)
+next_can_zero (void *nxdata)
 {
   struct b_conn *b_conn = nxdata;
-  return b_conn->b->pread (b_conn->b, b_conn->conn, buf, count, offset, 0);
+  return b_conn->b->can_zero (b_conn->b, b_conn->conn);
 }
 
 static int
-next_pwrite (void *nxdata, const void *buf, uint32_t count, uint64_t offset)
+next_can_fua (void *nxdata)
 {
   struct b_conn *b_conn = nxdata;
-  return b_conn->b->pwrite (b_conn->b, b_conn->conn, buf, count, offset, 0);
+  return b_conn->b->can_fua (b_conn->b, b_conn->conn);
 }
 
 static int
-next_flush (void *nxdata)
+next_pread (void *nxdata, void *buf, uint32_t count, uint64_t offset,
+            uint32_t flags, int *err)
 {
   struct b_conn *b_conn = nxdata;
-  return b_conn->b->flush (b_conn->b, b_conn->conn, 0);
+  return b_conn->b->pread (b_conn->b, b_conn->conn, buf, count, offset, flags,
+                           err);
 }
 
 static int
-next_trim (void *nxdata, uint32_t count, uint64_t offset)
+next_pwrite (void *nxdata, const void *buf, uint32_t count, uint64_t offset,
+             uint32_t flags, int *err)
 {
   struct b_conn *b_conn = nxdata;
-  return b_conn->b->trim (b_conn->b, b_conn->conn, count, offset, 0);
+  return b_conn->b->pwrite (b_conn->b, b_conn->conn, buf, count, offset, flags,
+                            err);
 }
 
 static int
-next_zero (void *nxdata, uint32_t count, uint64_t offset, int may_trim)
+next_flush (void *nxdata, uint32_t flags, int *err)
 {
   struct b_conn *b_conn = nxdata;
-  uint32_t f = 0;
+  return b_conn->b->flush (b_conn->b, b_conn->conn, flags, err);
+}
 
-  if (may_trim)
-    f |= NBDKIT_FLAG_MAY_TRIM;
+static int
+next_trim (void *nxdata, uint32_t count, uint64_t offset, uint32_t flags,
+           int *err)
+{
+  struct b_conn *b_conn = nxdata;
+  return b_conn->b->trim (b_conn->b, b_conn->conn, count, offset, flags, err);
+}
 
-  return b_conn->b->zero (b_conn->b, b_conn->conn, count, offset, f);
+static int
+next_zero (void *nxdata, uint32_t count, uint64_t offset, uint32_t flags,
+           int *err)
+{
+  struct b_conn *b_conn = nxdata;
+  return b_conn->b->zero (b_conn->b, b_conn->conn, count, offset, flags, err);
 }
 
 static struct nbdkit_next_ops next_ops = {
@@ -343,6 +349,8 @@ static struct nbdkit_next_ops next_ops = {
   .can_flush = next_can_flush,
   .is_rotational = next_is_rotational,
   .can_trim = next_can_trim,
+  .can_zero = next_can_zero,
+  .can_fua = next_can_fua,
   .pread = next_pread,
   .pwrite = next_pwrite,
   .flush = next_flush,
@@ -357,12 +365,19 @@ filter_prepare (struct backend *b, struct connection *conn)
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
 
-  debug ("prepare");
+  debug ("%s: prepare", f->name);
 
-  if (f->filter.prepare)
-    return f->filter.prepare (&next_ops, &nxdata, handle);
-  else
-    return f->backend.next->prepare (f->backend.next, conn);
+  /* Call these in order starting from the filter closest to the
+   * plugin.
+   */
+  if (f->backend.next->prepare (f->backend.next, conn) == -1)
+    return -1;
+
+  if (f->filter.prepare &&
+      f->filter.prepare (&next_ops, &nxdata, handle) == -1)
+    return -1;
+
+  return 0;
 }
 
 static int
@@ -372,12 +387,16 @@ filter_finalize (struct backend *b, struct connection *conn)
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
 
-  debug ("finalize");
+  debug ("%s: finalize", f->name);
 
-  if (f->filter.finalize)
-    return f->filter.finalize (&next_ops, &nxdata, handle);
-  else
-    return f->backend.next->finalize (f->backend.next, conn);
+  /* Call these in reverse order to .prepare above, starting from the
+   * filter furthest away from the plugin.
+   */
+  if (f->filter.finalize &&
+      f->filter.finalize (&next_ops, &nxdata, handle) == -1)
+    return -1;
+
+  return f->backend.next->finalize (f->backend.next, conn);
 }
 
 static int64_t
@@ -387,7 +406,7 @@ filter_get_size (struct backend *b, struct connection *conn)
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
 
-  debug ("get_size");
+  debug ("%s: get_size", f->name);
 
   if (f->filter.get_size)
     return f->filter.get_size (&next_ops, &nxdata, handle);
@@ -402,7 +421,7 @@ filter_can_write (struct backend *b, struct connection *conn)
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
 
-  debug ("can_write");
+  debug ("%s: can_write", f->name);
 
   if (f->filter.can_write)
     return f->filter.can_write (&next_ops, &nxdata, handle);
@@ -417,7 +436,7 @@ filter_can_flush (struct backend *b, struct connection *conn)
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
 
-  debug ("can_flush");
+  debug ("%s: can_flush", f->name);
 
   if (f->filter.can_flush)
     return f->filter.can_flush (&next_ops, &nxdata, handle);
@@ -432,7 +451,7 @@ filter_is_rotational (struct backend *b, struct connection *conn)
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
 
-  debug ("is_rotational");
+  debug ("%s: is_rotational", f->name);
 
   if (f->filter.is_rotational)
     return f->filter.is_rotational (&next_ops, &nxdata, handle);
@@ -447,7 +466,7 @@ filter_can_trim (struct backend *b, struct connection *conn)
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
 
-  debug ("can_trim");
+  debug ("%s: can_trim", f->name);
 
   if (f->filter.can_trim)
     return f->filter.can_trim (&next_ops, &nxdata, handle);
@@ -455,10 +474,40 @@ filter_can_trim (struct backend *b, struct connection *conn)
     return f->backend.next->can_trim (f->backend.next, conn);
 }
 
+static int
+filter_can_zero (struct backend *b, struct connection *conn)
+{
+  struct backend_filter *f = container_of (b, struct backend_filter, backend);
+  void *handle = connection_get_handle (conn, f->backend.i);
+  struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
+
+  debug ("%s: can_zero", f->name);
+
+  if (f->filter.can_zero)
+    return f->filter.can_zero (&next_ops, &nxdata, handle);
+  else
+    return f->backend.next->can_zero (f->backend.next, conn);
+}
+
+static int
+filter_can_fua (struct backend *b, struct connection *conn)
+{
+  struct backend_filter *f = container_of (b, struct backend_filter, backend);
+  void *handle = connection_get_handle (conn, f->backend.i);
+  struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
+
+  debug ("%s: can_fua", f->name);
+
+  if (f->filter.can_fua)
+    return f->filter.can_fua (&next_ops, &nxdata, handle);
+  else
+    return f->backend.next->can_fua (f->backend.next, conn);
+}
+
 static int
 filter_pread (struct backend *b, struct connection *conn,
               void *buf, uint32_t count, uint64_t offset,
-              uint32_t flags)
+              uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
   void *handle = connection_get_handle (conn, f->backend.i);
@@ -466,41 +515,42 @@ filter_pread (struct backend *b, struct connection *conn,
 
   assert (flags == 0);
 
-  debug ("pread count=%" PRIu32 " offset=%" PRIu64, count, offset);
+  debug ("%s: pread count=%" PRIu32 " offset=%" PRIu64 " flags=0x%" PRIx32,
+         f->name, count, offset, flags);
 
   if (f->filter.pread)
     return f->filter.pread (&next_ops, &nxdata, handle,
-                            buf, count, offset);
+                            buf, count, offset, flags, err);
   else
     return f->backend.next->pread (f->backend.next, conn,
-                                   buf, count, offset, flags);
+                                   buf, count, offset, flags, err);
 }
 
 static int
 filter_pwrite (struct backend *b, struct connection *conn,
                const void *buf, uint32_t count, uint64_t offset,
-               uint32_t flags)
+               uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
-  bool fua = flags & NBDKIT_FLAG_FUA;
 
   assert (!(flags & ~NBDKIT_FLAG_FUA));
 
-  debug ("pwrite count=%" PRIu32 " offset=%" PRIu64 " fua=%d",
-         count, offset, fua);
+  debug ("%s: pwrite count=%" PRIu32 " offset=%" PRIu64 " flags=0x%" PRIx32,
+         f->name, count, offset, flags);
 
   if (f->filter.pwrite)
     return f->filter.pwrite (&next_ops, &nxdata, handle,
-                             buf, count, offset);
+                             buf, count, offset, flags, err);
   else
     return f->backend.next->pwrite (f->backend.next, conn,
-                                    buf, count, offset, flags);
+                                    buf, count, offset, flags, err);
 }
 
 static int
-filter_flush (struct backend *b, struct connection *conn, uint32_t flags)
+filter_flush (struct backend *b, struct connection *conn, uint32_t flags,
+              int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
   void *handle = connection_get_handle (conn, f->backend.i);
@@ -508,18 +558,18 @@ filter_flush (struct backend *b, struct connection *conn, uint32_t flags)
 
   assert (flags == 0);
 
-  debug ("flush");
+  debug ("%s: flush flags=0x%" PRIx32, f->name, flags);
 
   if (f->filter.flush)
-    return f->filter.flush (&next_ops, &nxdata, handle);
+    return f->filter.flush (&next_ops, &nxdata, handle, flags, err);
   else
-    return f->backend.next->flush (f->backend.next, conn, flags);
+    return f->backend.next->flush (f->backend.next, conn, flags, err);
 }
 
 static int
 filter_trim (struct backend *b, struct connection *conn,
              uint32_t count, uint64_t offset,
-             uint32_t flags)
+             uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
   void *handle = connection_get_handle (conn, f->backend.i);
@@ -527,34 +577,36 @@ filter_trim (struct backend *b, struct connection *conn,
 
   assert (flags == 0);
 
-  debug ("trim count=%" PRIu32 " offset=%" PRIu64, count, offset);
+  debug ("%s: trim count=%" PRIu32 " offset=%" PRIu64 " flags=0x%" PRIx32,
+         f->name, count, offset, flags);
 
   if (f->filter.trim)
-    return f->filter.trim (&next_ops, &nxdata, handle, count, offset);
+    return f->filter.trim (&next_ops, &nxdata, handle, count, offset, flags,
+                           err);
   else
-    return f->backend.next->trim (f->backend.next, conn, count, offset, flags);
+    return f->backend.next->trim (f->backend.next, conn, count, offset, flags,
+                                  err);
 }
 
 static int
 filter_zero (struct backend *b, struct connection *conn,
-             uint32_t count, uint64_t offset, uint32_t flags)
+             uint32_t count, uint64_t offset, uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
   void *handle = connection_get_handle (conn, f->backend.i);
   struct b_conn nxdata = { .b = f->backend.next, .conn = conn };
-  int may_trim = (flags & NBDKIT_FLAG_MAY_TRIM) != 0;
 
   assert (!(flags & ~(NBDKIT_FLAG_MAY_TRIM | NBDKIT_FLAG_FUA)));
 
-  debug ("zero count=%" PRIu32 " offset=%" PRIu64 " may_trim=%d",
-         count, offset, may_trim);
+  debug ("%s: zero count=%" PRIu32 " offset=%" PRIu64 " flags=0x%" PRIx32,
+         f->name, count, offset, flags);
 
   if (f->filter.zero)
     return f->filter.zero (&next_ops, &nxdata, handle,
-                           count, offset, may_trim);
+                           count, offset, flags, err);
   else
     return f->backend.next->zero (f->backend.next, conn,
-                                  count, offset, flags);
+                                  count, offset, flags, err);
 }
 
 static struct backend filter_functions = {
@@ -567,7 +619,6 @@ static struct backend filter_functions = {
   .dump_fields = filter_dump_fields,
   .config = filter_config,
   .config_complete = filter_config_complete,
-  .errno_is_preserved = plugin_errno_is_preserved,
   .open = filter_open,
   .prepare = filter_prepare,
   .finalize = filter_finalize,
@@ -577,6 +628,8 @@ static struct backend filter_functions = {
   .can_flush = filter_can_flush,
   .is_rotational = filter_is_rotational,
   .can_trim = filter_can_trim,
+  .can_zero = filter_can_zero,
+  .can_fua = filter_can_fua,
   .pread = filter_pread,
   .pwrite = filter_pwrite,
   .flush = filter_flush,
@@ -591,7 +644,7 @@ filter_register (struct backend *next, size_t index, const char *filename,
 {
   struct backend_filter *f;
   const struct nbdkit_filter *filter;
-  size_t i, len, size;
+  size_t i, len;
 
   f = calloc (1, sizeof *f);
   if (f == NULL) {
@@ -619,23 +672,17 @@ filter_register (struct backend *next, size_t index, const char *filename,
     exit (EXIT_FAILURE);
   }
 
-  /* Check for incompatible future versions. */
-  if (filter->_api_version != 1) {
+  /* We do not provide API or ABI guarantees for filters, other than
+   * the ABI position of _api_version that will let us diagnose
+   * mismatch when the API changes.
+   */
+  if (filter->_api_version != NBDKIT_FILTER_API_VERSION) {
     fprintf (stderr, "%s: %s: filter is incompatible with this version of nbdkit (_api_version = %d)\n",
              program_name, f->filename, filter->_api_version);
     exit (EXIT_FAILURE);
   }
 
-  /* Since the filter might be much older than the current version of
-   * nbdkit, only copy up to the self-declared _struct_size of the
-   * filter and zero out the rest.  If the filter is much newer then
-   * we'll only call the "old" fields.
-   */
-  size = sizeof f->filter;      /* our struct */
-  memset (&f->filter, 0, size);
-  if (size > filter->_struct_size)
-    size = filter->_struct_size;
-  memcpy (&f->filter, filter, size);
+  f->filter = *filter;
 
   /* Only filter.name is required. */
   if (f->filter.name == NULL) {
@@ -671,8 +718,11 @@ filter_register (struct backend *next, size_t index, const char *filename,
 
   debug ("registered filter %s (name %s)", f->filename, f->name);
 
+  /* Set debug flags before calling load. */
+  set_debug_flags (dl, f->name);
+
   /* Call the on-load callback if it exists. */
-  debug ("%s: load", f->filename);
+  debug ("%s: load", f->name);
   if (f->filter.load)
     f->filter.load ();
 
