@@ -1,4 +1,4 @@
-#!/bin/bash -
+#!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2018 Red Hat Inc.
 # All rights reserved.
@@ -32,6 +32,7 @@
 # SUCH DAMAGE.
 
 set -e
+set -x
 
 files="cache.img cache.sock cache.pid"
 rm -f $files
@@ -40,7 +41,7 @@ rm -f $files
 truncate -s 1G cache.img
 
 # Run nbdkit with the caching filter.
-nbdkit -P cache.pid -U cache.sock --filter cache file file=cache.img
+nbdkit -P cache.pid -U cache.sock --filter=cache file file=cache.img
 
 # We may have to wait a short time for the pid file to appear.
 for i in `seq 1 10`; do
@@ -60,6 +61,8 @@ pid="$(cat cache.pid)"
 cleanup ()
 {
     status=$?
+    trap '' INT QUIT TERM EXIT ERR
+    echo $0: cleanup: exit code $status
 
     kill $pid
     rm -f $files
@@ -69,7 +72,7 @@ cleanup ()
 trap cleanup INT QUIT TERM EXIT ERR
 
 # Open the overlay and perform some operations.
-guestfish --format=raw -a 'nbd://?socket=cache.sock' <<'EOF'
+guestfish --format=raw -a "nbd://?socket=$PWD/cache.sock" <<'EOF'
   run
   part-disk /dev/sda gpt
   mkfs ext4 /dev/sda1
