@@ -1,4 +1,4 @@
-#!/bin/bash -
+#!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2017 Red Hat Inc.
 # All rights reserved.
@@ -92,6 +92,8 @@ pid="$(cat tls.pid)"
 cleanup ()
 {
     status=$?
+    trap '' INT QUIT TERM EXIT ERR
+    echo $0: cleanup: exit code $status
 
     kill $pid
     rm -f tls.pid tls.out
